#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>
#include <assert.h>
#include <errno.h>
#include <queue.h>
#include <workers.h>

// lunghezza massima per una singola lettura di linea con fgets
#define DEFAULT_MAX_LINE 2048
// numero di threads worker di default se non e' specificata l'opzione -t 
#define DEFAULT_NUM_WORKERS 4

// -------------- tipi di dato in ingresso ed in uscita a/da un worker ------------
typedef char* input_t;     // tipo di messaggio della coda di input
typedef struct output_t {  // tipo di messaggio della coda di output -- risposte
    char *filename;
    long numlines;
    long numwords;
} output_t;


// --------------- variabili globali ----------------------------------------------

// queste variabili sono definite nella libc e sono settate dalla getopt
extern int optind; 
extern char *optarg;
// flags
int countline = 0;  // 1 se devo contare le linee
int countword = 0;  // 1 se devo contare le parole
// lunghezza massima letta da una fgets -- puo' essere settata dal parametro -m
long max_line = DEFAULT_MAX_LINE;

// ---------------- funzioni di utilita' -----------------------------------------

// ritorna il numero di parole contenute nella stringa buffer passata come argomento
static int wcount(const char buf[]) {
    char *p = (char*)&buf[0];
    int count = 0, flag=0;
    // togliamo tutti gli spazi ed i caratteri di controllo
    // iniziali in modo da iniziare con un carattere
    while(*p != '\0' && (iscntrl(*p) || isspace(*p))) ++p;

    while(*p != '\0') { 
	if (isspace(*p)) {  // se ho uno spazio e se il carattere precedente 
	                    // non e' ne uno spazio ne un carattere di controllo....
	    if ( !isspace(*(p-1)) && !iscntrl(*(p-1)) ) { 
		++count;
		flag = 0;
	    }
	} else flag=1;  // ho iniziato una nuova parola
	++p;
    }
    return (flag) ? count+1:count;
}

static void usage(char *argv[]) {
	fprintf(stderr, "usa: %s [-l -w -m <num> -t <ths> ] <nome-file> [<nome-file> ...]\n", argv[0]);
	fprintf(stderr, " -l conta il numero di linee\n");
	fprintf(stderr, " -w conta il numero di parole\n");
	fprintf(stderr, " -m <num> setta la lunghezza massima di una linea a <num> [default: %d]\n", DEFAULT_MAX_LINE);
	fprintf(stderr, " -t <ths> setta il numero di threads a <ths> [default: %d]\n", DEFAULT_NUM_WORKERS); 
}

// -------------- funzione eseguita dal thread worker -----------------------------
// riceve in ingresso un nome di file la cui esistenza e' stata controllata
// ritorna il numero di linee e/o di parole del file attraverso u
void *Conta(void *indata) {
    assert(indata);
    char *filename = (char *)indata;
    
    // alloco il buffer per la fgets
    char *line = malloc(max_line*sizeof(char)); // buffer
    // preparo il messaggio di risposta
    output_t *res = (output_t*)malloc(sizeof(output_t));
    if (!line || !res) {
	fprintf(stderr, "Error allocating memory\n");
	if (line) free(line);
	if (res)  free(res);
	exit(EXIT_FAILURE);
    }
    // messaggio di risposta di default
    res->filename = filename;
    res->numlines = -1;
    res->numwords = -1;

    FILE *fp;
    if ((fp = fopen(filename,"r")) == NULL) {
	fprintf(stderr, "Error opening file %s\n", filename);
	perror("fopen");	    
	goto exit; 
    }

    long numlines=0, numwords=0;
    while(fgets(line, max_line, fp) != NULL) { 
	if (line[strlen(line)-1] == '\n') ++numlines;
	if (countword) numwords += wcount(line);
    }

    fclose(fp);

    res->filename = filename;
    res->numlines  = countline?numlines:-1;
    res->numwords  = countword?numwords:-1;

 exit:
    free(line);
    return res;
}
// ----------------------------------------------------------------------------


int main(int argc, char *argv[]) {
    if (argc == 1) {
	usage(argv);
	exit(EXIT_FAILURE);
    }
    int nth = DEFAULT_NUM_WORKERS;
    char opt;
    // parsing della command line
    while((opt = getopt(argc, argv, "lwm:t:")) != -1) {
	switch(opt) {
	case 'l': {
	    countline = 1;
	} break;
	case 'w': {
	    countword = 1;
	} break;
	case 'm': {
	    max_line = strtol(optarg, NULL, 10);
	} break;
	case 't': {
	    nth = strtol(optarg, NULL, 10);
	} break;
	default: ;
	}
    }

    // se non ci sono opzioni considero sia -l che -w come abilitate
    if (countword + countline == 0) { countline=1; countword=1; }

    pthread_t    *th;
    threadArgs_t *thARGS;

    // vettore dei thread ids
    th     = malloc(nth*sizeof(pthread_t));
    // vettore degli argomenti per i threads
    thARGS = malloc(nth*sizeof(threadArgs_t));
    if (!th || !thARGS) {
	fprintf(stderr, "malloc fallita\n");
	exit(EXIT_FAILURE);
    }

    // creo la coda di ingresso dei thread workers
    Queue_t *inQ = initQueue();
    if (!inQ) {
	fprintf(stderr, "initQueue fallita (coda di ingresso)\n");
	exit(errno);
    }
    // creo la coda di uscita  dei thread workers
    Queue_t *outQ = initQueue();
    if (!outQ) {
	fprintf(stderr, "initQueue fallita (coda di uscita)\n");
	exit(errno);
    }

    // preparo i parametri per i threads
    for(int i=0;i<nth; ++i) {	
	thARGS[i].thid = i;
	thARGS[i].inQ  = inQ;
	thARGS[i].outQ = outQ;
	thARGS[i].fun  = Conta;
	thARGS[i].EOS  = NULL;  // messaggio di terminazione
    }

    // creazione dei threads
    for(int i=0;i<nth; ++i)
	if (pthread_create(&th[i], NULL, Worker, &thARGS[i]) != 0) {
	    fprintf(stderr, "pthread_create failed\n");
	    exit(EXIT_FAILURE);
	}


    // per ogni file in ingresso, controllo che il file esista e sia un file 
    // regolare quindi inserisco il messaggio nella coda
    while(argv[optind] != NULL) {
	struct stat statbuf;
	char *filename = argv[optind];
	++optind;
	if (stat(filename, &statbuf)==-1) {
	    fprintf(stderr, "Errore, il file %s non esiste\n", filename);
	    perror("stat");	    
	    continue;
	}
	// se e' un file regolare lo considero altrimenti no
	if (S_ISREG(statbuf.st_mode)) push(inQ, filename); 

    }

    // invio il messaggio di terminazione a tutti i threads
    for (int i=0;i<nth;++i) push(inQ, thARGS[i].EOS);

    // raccolgo tutti i risultati ed aspetto che tutti i threads escano
    int neos = 0; // numero di messaggi di terminazione
    long total_lines=0, total_words=0;
    while(neos < nth) {
	output_t *res = (output_t*)pop(outQ);
	if (res == NULL) { neos++; continue;}
	if (res->numlines == -1 && res->numwords == -1) {
	    fprintf(stdout, " Errore nell'accesso al file %s\n", res->filename);
	    continue;
	}
	if (countline && !countword) { 
	    printf("  %10ld %s\n", res->numlines, res->filename);
	    total_lines+=res->numlines;
	}
	if (!countline && countword) {
	    printf("  %10ld %s\n", res->numwords, res->filename);
	    total_words+=res->numwords;
	}
	if (countline && countword)  {
	    printf("  %10ld  %10ld %s\n", res->numlines, res->numwords, res->filename);
	    total_lines+=res->numlines;
	    total_words+=res->numwords;
	}
	free(res);
    }
    printf("  ----------  ----------\n");
    if (countline && !countword) printf("  %10ld\n", total_lines);
    if (!countline && countword) printf("  %10ld\n", total_words);
    if (countline && countword)  printf("  %10ld  %10ld %s\n", total_lines, total_words, "total");

    for(int i=0;i<nth; ++i) pthread_join(th[i], NULL);
    free(th);
    free(thARGS);
    deleteQueue(inQ);
    deleteQueue(outQ);
    return 0;
}
